# django-salesforce
#
# by Hyneck Cernoch and Phil Christensen
# See LICENSE.md for details
#

"""
Multi-database support for the Django admin.
Not important yet - replaced by an empty stub.
"""
from django.contrib.admin.options import ModelAdmin as RoutedModelAdmin  # NOQA pylint:disable=unused-import
