# minimal dynamic metadata are specified here. Static metadata are in "pyproject.toml".
import os
import setuptools

sf_development = os.path.isfile('.djsf_development')
setuptools.setup(
    install_requires=[
        'django>=2.2' + (',<6.0' if not sf_development else ''),
        'pytz>=2012c',
        'requests>=2.32.0',
    ],
)
